# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
from typing import Iterable, Tuple, Callable, Dict
from collections import defaultdict
import functools
import threading
import dataclasses

try:
    import mido  # type: ignore
except ImportError as e:
    raise ImportError(
        "MIDI (joystick) support not installed, install optional dependencies with yakut[joystick]"
    ) from e
import yakut
from . import Controller, Sample, ControllerError, ControllerNotFoundError


_CH_MAX = 127


@dataclasses.dataclass()
class ButtonState:
    down: bool = False
    count: int = 0

    def update(self, down: bool) -> None:
        self.down = down
        self.count += down


class MIDIController(Controller):
    """
    Interface for standard MIDI HID controllers/faders.
    """

    def __init__(self, name: str) -> None:
        self._lock = threading.RLock()
        self._update_hook: Callable[[], None] = lambda: None
        self._analog: Dict[int, float] = {}
        self._buttons: Dict[int, ButtonState] = defaultdict(ButtonState)
        try:
            # noinspection PyUnresolvedReferences
            self._port: mido.ports.BaseInput = mido.open_input(name, callback=self._callback)
        except OSError as ex:
            raise ControllerNotFoundError(f"Cannot open MIDI port {name!r}") from ex
        except Exception as ex:
            raise ControllerError(str(ex)) from ex

        self._name = self._port.name or "?"
        _logger.info("%s: MIDI controller initialized: %r", self, name)

    @property
    def name(self) -> str:
        return self._name

    def sample(self) -> Sample:
        with self._lock:
            if self._port.closed:
                raise ControllerNotFoundError("MIDI port is closed")
            return Sample(
                axis=defaultdict(float, self._analog),
                button=defaultdict(bool, ((axis, state.down) for axis, state in self._buttons.items())),
                toggle=defaultdict(bool, ((axis, state.count % 2 != 0) for axis, state in self._buttons.items())),
            )

    def set_update_hook(self, hook: Callable[[], None]) -> None:
        self._update_hook = hook

    def close(self) -> None:
        with self._lock:
            self._port.close()

    def _handle_control_change(self, channel: int, control: int, value: int) -> None:
        # Accept all channels.
        _ = channel
        # From the MIDI 1.0 Spec:
        #   All controller number assignments are designated by agreement between the MMA and JMSC. The numbers listed
        #   in Table III are specified for standard musical instrument applications. However, many non-musical devices
        #   which implement MIDI, such as lighting controllers, may use designated controller numbers at their
        #   discretion.
        # Here, we treat all controls equally regardless of their fixed functions prescribed by the MIDI spec.
        axis = control
        # Automatically differentiate between buttons and sliders by evaluating the value: if only min or max
        # are seen, assume this is a button, otherwise, assume it is a slider. I am not yet sure how sensible
        # this is considering the edge case: when we first receive a slider data and it happens to be either
        # min or max, it will be initially categorized as a button, and any following update will cause this
        # logic to reclassify it as a slider. This may be confusing for users so maybe we should instead
        # treat every channel as both slider AND button?
        if axis in self._analog or value not in (0, _CH_MAX):
            self._analog[axis] = value / _CH_MAX
            self._buttons.pop(axis, None)
        else:
            self._buttons[axis].update(value > (_CH_MAX // 2))

    # noinspection PyUnresolvedReferences
    def _callback(self, msg: mido.Message) -> None:
        try:
            with self._lock:
                _logger.debug("%s: MIDI message: %s", self, msg)
                if msg.type == "control_change":
                    self._handle_control_change(msg.channel, msg.control, msg.value)

            self._update_hook()
        except Exception as ex:  # pylint: disable=broad-except
            _logger.exception("%s: MIDI event handler failure: %s", self, ex)  # pragma: no cover

    @staticmethod
    def list_controllers() -> Iterable[Tuple[str, Callable[[], Controller]]]:
        # noinspection PyUnresolvedReferences
        for name in mido.get_input_names():
            yield name, functools.partial(MIDIController, name)


_logger = yakut.get_logger(__name__)



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import io
from typing import Any, TextIO
import decimal
import ruamel.yaml


class Dumper:
    """
    YAML generation facade.
    Natively represents decimal.Decimal as floats in the output.
    """

    def __init__(self, explicit_start: bool = False):
        # We need to use the roundtrip representer to retain ordering of mappings, which is important for usability.
        self._impl = ruamel.yaml.YAML(typ="rt")
        # noinspection PyTypeHints
        self._impl.explicit_start = explicit_start
        self._impl.default_flow_style = None  # Choose between block/inline automatically
        self._impl.width = 2 ** 31  # Unlimited width

    def dump(self, data: Any, stream: TextIO) -> None:
        self._impl.dump(data, stream)

    def dumps(self, data: Any) -> str:
        s = io.StringIO()
        self.dump(data, s)
        return s.getvalue()


def _represent_decimal(self: ruamel.yaml.BaseRepresenter, data: decimal.Decimal) -> ruamel.yaml.ScalarNode:
    if data.is_finite():
        s = str(_POINT_ZERO_DECIMAL + data)  # The zero addition is to force float-like string representation
    elif data.is_nan():
        s = ".nan"
    elif data.is_infinite():
        s = ".inf" if data > 0 else "-.inf"
    else:
        assert False
    return self.represent_scalar("tag:yaml.org,2002:float", s)


ruamel.yaml.add_representer(decimal.Decimal, _represent_decimal, representer=ruamel.yaml.RoundTripRepresenter)

_POINT_ZERO_DECIMAL = decimal.Decimal("0.0")


def _unittest_yaml() -> None:
    ref = Dumper(explicit_start=True).dumps(
        {
            "abc": decimal.Decimal("-inf"),
            "def": [decimal.Decimal("nan"), {"qaz": decimal.Decimal("789")}],
        }
    )
    assert (
        ref
        == """---
abc: -.inf
def:
- .nan
- {qaz: 789.0}
"""
    )



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
from typing import Dict, Any


NAME_SEP = "."
ITEM_SEP = " "


class EnvironmentVariableError(ValueError):
    pass


def encode(value: Any) -> bytes:
    """
    Converts the register value to binary environment variable value as defined by the UAVCAN Specification,
    the Register interface.

    >>> encode(123)
    b'123'
    >>> encode('hello world')
    b'hello world'
    >>> encode(b'hello')
    b'hello'
    >>> encode([True, False, True])
    b'1 0 1'
    >>> encode([60_000, 50_000])
    b'60000 50000'
    >>> encode(300_000)
    b'300000'
    >>> encode([2 ** 32, 0])
    b'4294967296 0'
    >>> encode(-10_000)
    b'-10000'
    >>> encode([-10_000, 40_000])
    b'-10000 40000'
    >>> encode([-(2 ** 31), 2 ** 31])
    b'-2147483648 2147483648'
    >>> encode(1.0)
    b'1.0'
    """
    if isinstance(value, str):
        return value.encode()
    if isinstance(value, bytes):
        return value
    if isinstance(value, (int, bool)):
        return encode(str(int(value)))
    if isinstance(value, float):
        return encode(str(value))
    if isinstance(value, list):
        return b" ".join(map(encode, value))
    raise EnvironmentVariableError(f"Cannot encode register value of type {type(value).__name__}")


def flatten_registers(spec: Dict[str, Any], prefix: str = "") -> Dict[str, Any]:
    """
    >>> flatten_registers({"FOO": "BAR", "a": {"b": 123, "c": [456, 789]}})  # doctest: +NORMALIZE_WHITESPACE
    {'FOO': 'BAR',
     'a.b': 123,
     'a.c': [456, 789]}
    """
    out: Dict[str, Any] = {}
    for k, v in spec.items():
        name = NAME_SEP.join((prefix, k)) if prefix else k
        if isinstance(v, dict):
            out.update(flatten_registers(v, name))
        else:
            out[name] = v
    return out



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import sys
import dataclasses
import functools
from typing import Optional, Iterable, Any
from numbers import Number
import enum
import click


# errors=ignore is necessary to handle Unicode support limitations on Windows.
_TEXT_STREAM = click.get_text_stream("stdout", errors="ignore")


def refresh_screen(contents: str) -> None:
    click.echo("\n")  # Mark the boundary between screens when redirecting to file.
    click.clear()
    sys.stdout.flush()  # Synchronize clear with the following output since it is buffered separately.
    click.echo(contents, file=_TEXT_STREAM)


class Color(enum.Enum):
    BLACK = enum.auto()
    RED = enum.auto()
    GREEN = enum.auto()
    YELLOW = enum.auto()
    BLUE = enum.auto()
    MAGENTA = enum.auto()
    CYAN = enum.auto()
    WHITE = enum.auto()


@dataclasses.dataclass(frozen=True)
class Style:
    fg: Optional[Color] = None
    bg: Optional[Color] = None
    salience: int = 0  # -1 dim, +1 bold, +2 blink


class TableRenderer:
    """
    This class may be slow to instantiate. Do not instantiate inside the drawing loop.
    """

    def __init__(self, column_widths: Iterable[int], *, separate_columns: bool) -> None:
        self._canvas = Canvas()
        self._column_widths = list(column_widths)
        self._column_offsets = [0]
        for cw in self._column_widths:
            self._column_offsets.append(self._column_offsets[-1] + cw + int(separate_columns))

    def set_cell(self, row: int, column: int, data: Any, *, style: Optional[Style] = None) -> None:
        if isinstance(data, (bool, Number)):
            data = str(data).rjust(self._column_widths[column])
        self._canvas.put(row, self._column_offsets[column], data, style=style)

    def __setitem__(self, key: tuple[int, int], value: Any) -> None:
        row, col = key
        if isinstance(value, tuple):
            data, style = value
        else:
            data, style = value, None
        self.set_cell(row, col, data, style=style)

    def flip_buffer(self) -> str:
        # Make all rows equal length.
        m_row, m_col = self._canvas.extent
        for row in range(m_row):
            self._canvas.put(row, m_col, "")
        return self._canvas.flip_buffer()


class Canvas:
    @dataclasses.dataclass(frozen=True)
    class _Block:
        column: int
        text: str
        style: Optional[Style]

    def __init__(self) -> None:
        self._rows: list[list[Canvas._Block]] = []

    @property
    def extent(self) -> tuple[int, int]:
        return len(self._rows), max((x.column + len(x.text)) for r in self._rows for x in r)

    def put(self, row: int, column: int, data: Any, *, style: Optional[Style] = None) -> int:
        while len(self._rows) <= row:
            self._rows.append([])
        text = str(data)
        bl = Canvas._Block(
            column=column,
            text=text,
            style=style,
        )
        self._rows[row].append(bl)
        return column + len(text)

    def flip_buffer(self) -> str:
        out = "\n".join(map(self._render_row, self._rows)) + click.style("", reset=True)
        self._rows = []
        return out

    def _render_row(self, ln: list[Canvas._Block]) -> str:
        col = 0
        out: list[str] = [self._begin_style(None)]
        for b in sorted(ln, key=lambda x: x.column):
            out.append(" " * (b.column - col))
            col = b.column
            out.append(self._begin_style(b.style))
            out.append(b.text)
            col += len(b.text)
        return "".join(out)

    @staticmethod
    @functools.lru_cache(None)
    def _begin_style(s: Optional[Style]) -> str:
        out = click.style("", reset=True)
        if s is not None:
            out += click.style(
                "",
                fg=(("bright_" * (s.salience > 0)) + s.fg.name.lower()) if s.fg else None,
                bg=s.bg.name.lower() if s.bg else None,
                dim=s.salience < 0,
                bold=s.salience > 0,
                blink=s.salience > 1,
                reset=False,
            )
        return out



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
from typing import Set, Any, Callable, Optional
import yakut
from yakut.controller import Controller, list_controllers, ControllerNotFoundError
from yakut.controller import Sample as Sample


class ControllerReader:
    """
    This class can be trivially updated to emit notifications when a controller input is received
    (i.e., when the user interacts with the controller like moving a slider or pressing a button).
    """

    def __init__(self) -> None:
        """
        The list of available controllers is sampled ONCE during initialization and held forever because:

        - This is a slow operation that involves much IO.
        - Controllers may be connected/disconnected at runtime, rendering their indices obsolete.
        """
        self._sources = [ControlInputSource(name, factory) for name, factory in list_controllers()]

    @property
    def active(self) -> Set[str]:
        """
        Names of currently active controllers (i.e., those that are read from).
        """
        return set(x.name for x in self._sources if x.active)

    def sample_and_hold(self) -> None:
        """
        Capture the state of all controllers nearly atomically.
        The updated state can be collected using :meth:`read` separately per controller.
        """
        for s in self._sources:
            s.sample_and_hold()

    def read(self, selector: Any) -> Sample:
        """
        TODO: support abbreviated case-insensitive names for persistence;
        e.g., ``xbox`` to match the first X-Box 360 Controller.

        :raises: :class:`ControllerNotFoundError`
        """
        try:
            selector = int(selector)
        except (ValueError, TypeError):  # pragma: no cover
            raise ControllerNotFoundError(f"Non-index controller selectors not yet supported: {selector!r}") from None
        try:
            src = self._sources[selector]
        except LookupError:
            raise ControllerNotFoundError(f"No controller at index {selector}") from None
        out = src.read()
        assert isinstance(out, Sample)
        return out

    def close(self) -> None:
        for s in self._sources:
            s.close()

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._sources})"


class ControlInputSource:
    def __init__(self, name: str, factory: Callable[[], Controller]) -> None:
        self.name = name
        self._factory = factory
        self._controller: Optional[Controller] = None
        self._last_sample: Optional[Sample] = None

    @property
    def active(self) -> bool:
        return self._controller is not None

    def sample_and_hold(self) -> None:
        if self._controller is not None:
            self._last_sample = self._controller.sample()

    def read(self) -> Sample:
        if self._controller is None:
            _logger.debug("%s: Connecting", self)
            self._controller = self._factory()
            self._last_sample = self._controller.sample()
        assert self._last_sample is not None
        return self._last_sample

    def close(self) -> None:
        if self._controller is not None:
            _logger.debug("%s: Closing", self)
            self._controller.close()

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.name!r})"


_logger = yakut.get_logger(__name__)


def _unittest_null() -> None:
    import pytest

    cr = ControllerReader()
    assert cr.active == set()
    cr.sample_and_hold()  # No-op because no controllers connected.

    assert cr.read("0")
    assert cr.active == {"null"}

    assert cr.read("0")  # Already connected
    assert cr.active == {"null"}

    with pytest.raises(ControllerNotFoundError):
        cr.read("999")
    assert cr.active == {"null"}

    cr.sample_and_hold()

    assert cr.read("0").axis == {}
    assert cr.read("0").button == {}
    assert cr.read("0").toggle == {}

    print(cr)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
from typing import Callable, Any, cast
from collections.abc import Mapping, Collection
import click

Formatter = Callable[[Any], str]
FormatterFactory = Callable[[], Formatter]


def formatter_factory_option(f: Callable[..., Any]) -> Callable[..., Any]:
    def validate(ctx: click.Context, param: object, value: str) -> FormatterFactory:
        _ = ctx, param
        try:
            return _FORMATTERS[value.upper()]
        except LookupError:
            raise click.BadParameter(f"Invalid format name: {value!r}") from None

    choices = list(_FORMATTERS.keys())
    default = choices[0]
    doc = f"""
The format of data printed into stdout.
This option is only relevant for commands that generate structured outputs, like pub or call; other commands ignore it.

The final representation of the output data is constructed from an intermediate "builtin-based" representation,
which is a simplified form that is stripped of the detailed DSDL type information, like JSON.
For more info please read the PyUAVCAN documentation on builtin-based representations.

YAML separates objects with `---`.

JSON and TSV (tab separated values) keep exactly one object per line.

TSV is intended for use with third-party software
such as computer algebra systems or spreadsheet processors.

TSVH is just TSV with the header included.

TSVFC is TSVH with extra column for curly braces and square brackets. These are extra format columns that help the
reader understand the structure of the data without looking at the headers.
"""
    f = click.option(
        "--format",
        "-F",
        "formatter_factory",
        envvar="YAKUT_FORMAT",
        type=click.Choice(choices, case_sensitive=False),
        callback=validate,
        default=default,
        show_default=True,
        help=doc,
    )(f)
    return f


def _make_yaml_formatter() -> Formatter:
    from yakut.yaml import Dumper

    dumper = Dumper(explicit_start=True)
    return dumper.dumps


def _make_json_formatter() -> Formatter:
    # We prefer simplejson over the standard json because the native json lacks important capabilities:
    #  - simplejson preserves dict ordering, which is very important for UX.
    #  - simplejson supports Decimal.
    import simplejson as json  # type: ignore

    return lambda data: cast(str, json.dumps(data, ensure_ascii=False, separators=(",", ":")))


def _insert_format_specifier(
    items: list[tuple[str, Any]],
    key: str,
    instance: Collection[Any] | Mapping[Any, Any],
    is_start: bool = True,
) -> None:
    is_list = isinstance(instance, Collection) and not isinstance(instance, str)
    is_dictionary = isinstance(instance, Mapping)
    if is_start:
        if is_dictionary:
            items.append((key + "{", "{"))
        elif is_list:
            items.append((key + "[", "["))
    else:
        if is_dictionary:
            items.append((key + "}", "}"))
        elif is_list:
            items.append((key + "]", "]"))


def _flatten_start(
    d: dict[Any, Any] | Collection[Any],
    parent_key: str = "",
    sep: str = ".",
    do_put_format_specifiers: bool = False,
) -> dict[str, Any]:
    def flatten(d: dict[Any, Any] | Collection[Any], parent_key: str = "") -> dict[str, Any]:
        def add_item(items: list[tuple[str, Any]], new_key: str, v: Mapping[Any, Any] | Collection[Any]) -> None:
            if do_put_format_specifiers:
                _insert_format_specifier(items, new_key, v)
            if isinstance(v, Mapping) or (isinstance(v, Collection) and not isinstance(v, str)):
                for_extension = flatten(v, new_key)
                if for_extension is not None:
                    # noinspection PyTypeChecker
                    items.extend(for_extension.items())
            else:
                items.append((new_key, v))
            if do_put_format_specifiers:
                _insert_format_specifier(items, new_key, v, is_start=False)

        if isinstance(d, Mapping):
            items: list[tuple[str, Any]] = []
            for k, v in d.items():
                new_key = parent_key + sep + str(k) if parent_key else str(k)
                add_item(items, new_key, v)
            return dict(items)
        if isinstance(d, Collection) and not isinstance(d, str):
            items = []
            for i, v in enumerate(d):
                new_key = parent_key + sep + f"[{i}]" if parent_key else str(f"[{i}]")
                add_item(items, new_key, v)
            return dict(items)
        return {}

    return flatten(d, parent_key)


def _make_tsv_formatter() -> Formatter:
    def tsv_format_function(data: dict[Any, Any]) -> str:
        return "\t".join([str(v) for k, v in _flatten_start(data).items()])

    return tsv_format_function


def _make_tsvh_formatter_factory(do_put_format_specifiers: bool = False) -> Callable[[], Formatter]:
    def _make_tsvh_formatter() -> Formatter:
        is_first_time = True

        def tsv_format_function_with_header(data: dict[Any, Any]) -> str:
            nonlocal is_first_time
            if is_first_time:
                is_first_time = False
                return (
                    "\t".join(
                        [
                            str(k)
                            for k, v in _flatten_start(data, do_put_format_specifiers=do_put_format_specifiers).items()
                        ]
                    )
                    + "\n"
                    + "\t".join(
                        [
                            str(v)
                            for k, v in _flatten_start(data, do_put_format_specifiers=do_put_format_specifiers).items()
                        ]
                    )
                )
            return "\t".join(
                [str(v) for k, v in _flatten_start(data, do_put_format_specifiers=do_put_format_specifiers).items()]
            )

        return tsv_format_function_with_header

    return _make_tsvh_formatter


_FORMATTERS = {
    "YAML": _make_yaml_formatter,
    "JSON": _make_json_formatter,
    "TSV": _make_tsv_formatter,
    "TSVH": _make_tsvh_formatter_factory(do_put_format_specifiers=False),
    "TSVFC": _make_tsvh_formatter_factory(do_put_format_specifiers=True),
}


def _unittest_formatter() -> None:
    obj = {
        2345: {
            "abc": {
                "def": [123, 456],
            },
            "ghi": 789,
        }
    }
    assert (
        _FORMATTERS["YAML"]()(obj)
        == """---
2345:
  abc:
    def: [123, 456]
  ghi: 789
"""
    )
    assert _FORMATTERS["JSON"]()(obj) == '{"2345":{"abc":{"def":[123,456]},"ghi":789}}'
    assert _FORMATTERS["TSV"]()(obj) == "123\t456\t789"
    tsvh_formatter = _FORMATTERS["TSVH"]()
    # first time should include a header
    assert tsvh_formatter(obj) == "2345.abc.def.[0]\t2345.abc.def.[1]\t2345.ghi\n123\t456\t789"
    # subsequent calls shouldn't include a header
    assert tsvh_formatter(obj) == "123\t456\t789"
    from decimal import Decimal
    from math import nan

    obj = {
        142: {
            "_metadata_": {
                "timestamp": {"system": Decimal("1640611164.396007"), "monotonic": Decimal("4765.594161")},
                "priority": "nominal",
                "transfer_id": 28,
                "source_node_id": 21,
            },
            "timestamp": {"microsecond": 309697890},
            "value": {
                "kinematics": {
                    "angular_position": {"radian": nan},
                    "angular_velocity": {"radian_per_second": 0.0},
                    "angular_acceleration": {"radian_per_second_per_second": 0.0},
                },
                "torque": {"newton_meter": nan},
            },
        }
    }
    tsvfc_formatter = _FORMATTERS["TSVFC"]()
    assert (
        tsvfc_formatter(obj) == "142{	142._metadata_{	142._metadata_.timestamp{"
        "	142._metadata_.timestamp.system	142._metadata_.timestamp.monotonic"
        "	142._metadata_.timestamp}	142._metadata_.priority	142._metadata_.transfer_id"
        "	142._metadata_.source_node_id	142._metadata_}	142.timestamp{	142.timestamp.microsecond"
        "	142.timestamp}	142.value{	142.value.kinematics{	142.value.kinematics.angular_position{"
        "	142.value.kinematics.angular_position.radian	142.value.kinematics.angular_position}"
        "	142.value.kinematics.angular_velocity{	142.value.kinematics.angular_velocity.radian_per_second"
        "	142.value.kinematics.angular_velocity}	142.value.kinematics.angular_acceleration{"
        "	142.value.kinematics.angular_acceleration.radian_per_second_per_second"
        "	142.value.kinematics.angular_acceleration}	142.value.kinematics}	142.value.torque{"
        "	142.value.torque.newton_meter	142.value.torque}	142.value}	142}\n{	{	{"
        "	1640611164.396007	4765.594161	}	nominal	28	21	}"
        "	{	309697890	}	{	{	{	nan	}	{	0.0	}"
        "	{	0.0	}	}	{	nan	}	}	}"
    )
    assert _FORMATTERS["TSV"]()(obj) == "1640611164.396007\t4765.594161\tnominal\t28\t21\t309697890\tnan\t0.0\t0.0\tnan"

    assert (
        _FORMATTERS["TSVH"]()(obj)
        == "142._metadata_.timestamp.system\t142._metadata_.timestamp.monotonic\t142._metadata_.priority\t"
        "142._metadata_.transfer_id\t142._metadata_.source_node_id\t142.timestamp.microsecond\t"
        "142.value.kinematics.angular_position.radian"
        "\t142.value.kinematics.angular_velocity.radian_per_second\t142.value.kinematics.angular_acceleration."
        "radian_per_second_per_second\t142.value.torque.newton_meter"
        "\n1640611164.396007\t4765.594161\tnominal\t2"
        "8\t21\t309697890\tnan\t0.0\t0.0\tnan"
    )



# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import os
import sys
from typing import TYPE_CHECKING, Iterable, Optional, Any, Callable, Awaitable
import logging
from pathlib import Path
from shutil import get_terminal_size
import click
import yakut
from yakut.param.transport import transport_factory_option, TransportFactory, Transport
from yakut.param.formatter import formatter_factory_option, FormatterFactory, Formatter
from yakut.param.node import node_factory_option, NodeFactory

if TYPE_CHECKING:
    import pyuavcan.application


def get_logger(name: str) -> logging.Logger:
    """
    This is a trivial wrapper over :func:`logging.getLogger` that removes private components from the logger name.
    For example, ``yakut.cmd.file_server._cmd`` becomes ``yakut.cmd.file_server`` (private submodule hidden).
    Also, double underscores are removed.
    All this is done to make the log messages appear nicer, since this is important for a CLI tool.
    """
    return logging.getLogger(name.replace("__", "").split("._", 1)[0])


_logger = get_logger("yakut")
# Some of the integration tests may parse the logs expecting its lines to follow this format.
# If you change this, you may break these tests.
_LOG_FORMAT = "%(asctime)s %(process)07d %(levelname)-3.3s %(name)s: %(message)s"
logging.basicConfig(format=_LOG_FORMAT)  # Using the default log level; it will be overridden later.


class Purser:
    def __init__(
        self,
        paths: Iterable[str | Path],
        formatter_factory: FormatterFactory,
        transport_factory: TransportFactory,
        node_factory: NodeFactory,
    ) -> None:
        self._paths = list(Path(x) for x in paths)
        self._f_formatter = formatter_factory
        self._f_transport = transport_factory
        self._f_node = node_factory

        self._registry: Optional[pyuavcan.application.register.Registry] = None
        self._transport: Optional[Transport] = None
        self._node: Optional["pyuavcan.application.Node"] = None

    @property
    def paths(self) -> list[Path]:
        return list(self._paths)

    def make_formatter(self) -> Formatter:
        return self._f_formatter()

    def get_registry(self) -> pyuavcan.application.register.Registry:
        """
        Commands should never construct registry on their own!
        Doing so is likely to create divergent configurations that are not exposed via the Register Interface.
        Instead, use this factory: it will create a registry instance at the first invocation and then it will be
        shared with all components.

        :raises: :class:`ImportError` if the standard DSDL namespace ``uavcan`` is not available.
        """
        if self._registry is None:
            from pyuavcan.application import make_registry

            self._registry = make_registry()
        return self._registry

    def get_transport(self) -> Transport:
        if self._transport is None:  # pragma: no branch
            self._transport = self._f_transport()
        if self._transport is not None:
            return self._transport
        click.get_current_context().fail("Transport not configured, or the standard DSDL namespace is not compiled")

    def get_node(self, name_suffix: str, allow_anonymous: bool) -> "pyuavcan.application.Node":
        if self._node is None:  # pragma: no branch
            tr = self.get_transport()
            self._node = self._f_node(tr, name_suffix=name_suffix, allow_anonymous=allow_anonymous)
        return self._node


pass_purser = click.make_pass_decorator(Purser)


class AbbreviatedGroup(click.Group):
    def get_command(self, ctx: click.Context, cmd_name: str) -> Optional[click.Command]:
        cmd_name = cmd_name.replace("_", "-")
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        if len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail(f"Abbreviated command {cmd_name!r} is ambiguous. Possible matches: {list(matches)}")


_ENV_VAR_PATH = "YAKUT_PATH"


@click.command(
    cls=AbbreviatedGroup,
    context_settings={
        "max_content_width": get_terminal_size()[0],
        "auto_envvar_prefix": "YAKUT",  # Specified here, not in __main__.py, otherwise doesn't work when installed.
    },
)
@click.version_option(version=yakut.__version__)
@click.option("--verbose", "-v", count=True, help="Emit verbose log messages. Specify twice for extra verbosity.")
@click.option(
    "--path",
    "-P",
    multiple=True,
    type=click.Path(resolve_path=True),
    help=f"""
In order to use compiled DSDL namespaces,
the directories that contain compilation outputs need to be specified using this option.
The current working directory does not need to be specified explicitly.

Examples:

\b
    yakut  --path ../public_regulated_data_types  --path ~/my_namespaces  pub ...

\b
    export {_ENV_VAR_PATH}="../public_regulated_data_types:~/my_namespaces"
    yakut pub ...
""",
)
@formatter_factory_option
@transport_factory_option
@node_factory_option
@click.pass_context
def main(
    ctx: click.Context,
    verbose: int,
    path: tuple[str, ...],
    formatter_factory: FormatterFactory,
    transport_factory: TransportFactory,
    node_factory: NodeFactory,
) -> None:
    """
    \b
         __   __   _______   __   __   _______   _______   __   __
        |  | |  | /   _   ` |  | |  | /   ____| /   _   ` |  ` |  |
        |  | |  | |  |_|  | |  | |  | |  |      |  |_|  | |   `|  |
        |  |_|  | |   _   | `  `_/  / |  |____  |   _   | |  |`   |
        `_______/ |__| |__|  `_____/  `_______| |__| |__| |__| `__|
            |      |            |         |      |         |
        ----o------o------------o---------o------o---------o-------

    Yakut is a cross-platform command-line utility for diagnostics and management of UAVCAN networks.
    It is designed for use either directly by humans or from automation scripts.
    Ask questions at https://forum.uavcan.org

    Any long option can be provided via environment variable prefixed with `YAKUT_`
    such that an option `--foo-bar` for command `baz`, if not provided as a command-line argument,
    will be read from `YAKUT_BAZ_FOO_BAR`.

    Any command can be abbreviated arbitrarily as long as the resulting abridged name is not ambiguous.
    For example, `publish`, `publ` and `pub` are all valid and equivalent.
    """
    _configure_logging(verbose)  # This should be done in the first order to ensure that we log things correctly.

    path = (os.getcwd(), *path)
    _logger.debug("Path: %r", path)
    for p in path:
        sys.path.append(str(p))

    ctx.obj = Purser(
        paths=path,
        formatter_factory=formatter_factory,
        transport_factory=transport_factory,
        node_factory=node_factory,
    )


subcommand: Callable[..., Callable[..., Any]] = main.command  # type: ignore


def asynchronous(f: Callable[..., Awaitable[Any]]) -> Callable[..., Any]:
    import asyncio
    from functools import update_wrapper

    def proxy(*args: Any, **kwargs: Any) -> Any:
        return asyncio.run(f(*args, **kwargs))

    return update_wrapper(proxy, f)


def _configure_logging(verbosity_level: int) -> None:
    log_level = {
        0: logging.WARNING,
        1: logging.INFO,
        2: logging.DEBUG,
    }.get(verbosity_level or 0, logging.DEBUG)

    logging.root.setLevel(log_level)

    try:
        import coloredlogs  # type: ignore

        # The level spec applies to the handler, not the root logger! This is different from basicConfig().
        coloredlogs.install(level=log_level, fmt=_LOG_FORMAT)
    except Exception as ex:  # pylint: disable=broad-except
        _logger.exception("Could not set up coloredlogs: %r", ex)  # pragma: no cover



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
from typing import Iterable, Tuple, Callable
from collections import defaultdict
import yakut
from . import Controller, Sample


class NullController(Controller):
    """
    The null controller is useful if no real control inputs are needed.
    It always exists in exactly one instance and all its axes/buttons/etc. always read as zero/false.
    """

    NAME = "null"

    def __init__(self) -> None:
        _logger.info("%s: Initialized", self)

    @property
    def name(self) -> str:
        return NullController.NAME

    def sample(self) -> Sample:
        return Sample(axis=defaultdict(float), button=defaultdict(bool), toggle=defaultdict(bool))

    def set_update_hook(self, hook: Callable[[], None]) -> None:
        _logger.debug("%s: Update hook ignored because the null controller is never updated: %r", self, hook)

    def close(self) -> None:
        _logger.debug("%s: Closed (this is a no-op)", self)

    @staticmethod
    def list_controllers() -> Iterable[Tuple[str, Callable[[], Controller]]]:
        yield NullController.NAME, NullController


_logger = yakut.get_logger(__name__)



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Callable, Type, AbstractSet
import dataclasses
import math
import numpy as np
from numpy.typing import NDArray
import pyuavcan
from pyuavcan.transport import MessageDataSpecifier, ServiceDataSpecifier, Timestamp, AlienTransfer
import yakut
from ._iface import Iface

if TYPE_CHECKING:
    import uavcan.node
    import uavcan.node.port


@dataclasses.dataclass()
class PortSet:
    pub: AbstractSet[int] = dataclasses.field(default_factory=frozenset)
    sub: AbstractSet[int] = dataclasses.field(default_factory=frozenset)
    cln: AbstractSet[int] = dataclasses.field(default_factory=frozenset)
    srv: AbstractSet[int] = dataclasses.field(default_factory=frozenset)


@dataclasses.dataclass(frozen=True)
class NodeState:
    online: bool
    """
    Online means that the node is emitting any transfers whatsoever.
    """

    heartbeat: Optional[uavcan.node.Heartbeat_1_0]
    """
    An online node without a heartbeat is a zombie, which is an error condition because heartbeats are required
    for all nodes unconditionally.
    """
    info: Optional[uavcan.node.GetInfo_1_0.Response]

    ports: Optional[PortSet]
    """
    Defined only if the node keeps its uavcan.node.List publications up-to-date.
    """


class Avatar:  # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        iface: Iface,
        node_id: Optional[int],
        info: Optional[uavcan.node.GetInfo_1_0.Response] = None,
    ) -> None:
        import uavcan.node
        import uavcan.node.port

        self._node_id = node_id

        self._heartbeat: Optional[uavcan.node.Heartbeat_1_0] = None
        self._iface = iface
        self._info = info
        self._num_info_requests = 0

        self._ts_activity = -math.inf
        self._ts_heartbeat = -math.inf
        self._ts_port_list = -math.inf
        self._ts_info_request = -math.inf

        self._ports = PortSet()

        self._dispatch: dict[
            Type[pyuavcan.dsdl.FixedPortCompositeObject]
            | tuple[Type[pyuavcan.dsdl.FixedPortServiceObject], ServiceDataSpecifier.Role],
            Callable[[float, pyuavcan.dsdl.CompositeObject], None],
        ] = {
            (uavcan.node.GetInfo_1_0, ServiceDataSpecifier.Role.RESPONSE): self._on_info_response,
            uavcan.node.port.List_0_1: self._on_port_list,
            uavcan.node.Heartbeat_1_0: self._on_heartbeat,
        }

        self._iface.add_standard_subscription(uavcan.node.Heartbeat_1_0)
        self._iface.add_standard_subscription(uavcan.node.port.List_0_1)
        self._iface.add_trace_handler(self._on_trace)

    def _restart(self) -> None:
        self._info = None
        self._num_info_requests = 0
        self._ts_port_list = -math.inf

    def _on_info_response(self, ts: float, obj: pyuavcan.dsdl.CompositeObject) -> None:
        import uavcan.node

        _logger.info("%r: Received node info", self)
        assert isinstance(obj, uavcan.node.GetInfo_1_0.Response)
        _ = ts
        self._info = obj

    def _on_port_list(self, ts: float, obj: pyuavcan.dsdl.CompositeObject) -> None:
        import uavcan.node.port

        assert isinstance(obj, uavcan.node.port.List_0_1)
        self._ports.pub = expand_subjects(obj.publishers)
        self._ports.sub = expand_subjects(obj.subscribers)
        self._ports.cln = expand_mask(obj.clients.mask)
        self._ports.srv = expand_mask(obj.servers.mask)
        self._ts_port_list = ts

    def _on_heartbeat(self, ts: float, obj: pyuavcan.dsdl.CompositeObject) -> None:
        from uavcan.node import Heartbeat_1_0 as Heartbeat, GetInfo_1_0 as GetInfo

        assert isinstance(obj, Heartbeat)

        # We used to have a node-ID collision heuristic here that checked if the timestamp is oscillating back and
        # forth, as it would indicate that there are multiple nodes running on the same node-ID. While this
        # heuristic is correct, it is ineffective in practice because heartbeats of nodes with a lower uptime
        # would have lower transfer-ID values, which (unless the transport is cyclic-TID) would make the transfer
        # reassembler discard such new heartbeats from conflicting nodes as duplicates (already seen transfers).
        # It is therefore impossible to detect collisions at this layer (it is possible only below the transport
        # layer). Although it might *occasionally* work if the heartbeats are delayed or published irregularly.

        # Invalidate the node info if the uptime goes backwards or if we received a heartbeat after a long pause.
        restart = self._heartbeat and (
            (self._heartbeat.uptime > obj.uptime) or (ts - self._ts_heartbeat > Heartbeat.OFFLINE_TIMEOUT)
        )
        if restart:
            _logger.info("%r: Restart detected: %r", self, obj)
            self._restart()

        if not self._info and self._node_id is not None:
            timeout = 2 ** (self._num_info_requests + 2)
            if ts - self._ts_info_request >= timeout:
                _logger.debug("%r: Would request info; timeout=%.1f", self, timeout)
                self._num_info_requests += 1
                self._ts_info_request = ts
                self._iface.try_request(GetInfo, self._node_id, GetInfo.Request())

        self._heartbeat = obj
        self._ts_heartbeat = ts

    def _on_trace(self, ts: Timestamp, tr: AlienTransfer) -> None:
        from pyuavcan.dsdl import get_fixed_port_id

        own = tr.metadata.session_specifier.source_node_id == self._node_id
        if not own:
            return
        ds = tr.metadata.session_specifier.data_specifier
        self._ts_activity = float(ts.monotonic)

        # Snoop on transfers sent by our node. Even if we can't ask we can still learn things by listening.
        for ty, handler in self._dispatch.items():
            if isinstance(ty, tuple):
                ty, role = ty
                assert isinstance(ty, type) and isinstance(role, ServiceDataSpecifier.Role)
                if isinstance(ds, ServiceDataSpecifier) and ds.role == role and ds.service_id == get_fixed_port_id(ty):
                    rr = getattr(ty, role.name.capitalize())
                    assert issubclass(rr, pyuavcan.dsdl.CompositeObject)
                    obj = pyuavcan.dsdl.deserialize(rr, tr.fragmented_payload)
                    _logger.debug("%r: Service snoop: %r from %r", self, obj, tr)
                    if obj is not None:
                        handler(float(ts.monotonic), obj)
            elif isinstance(ty, type) and issubclass(ty, pyuavcan.dsdl.FixedPortCompositeObject):
                if isinstance(ds, MessageDataSpecifier) and ds.subject_id == get_fixed_port_id(ty):
                    obj = pyuavcan.dsdl.deserialize(ty, tr.fragmented_payload)
                    _logger.debug("%r: Message snoop: %r from %r", self, obj, tr)
                    if obj is not None:
                        handler(float(ts.monotonic), obj)
            else:
                assert False

    def update(self, ts: float) -> NodeState:
        from uavcan.node import Heartbeat_1_0 as Heartbeat
        from uavcan.node.port import List_0_1 as PortList

        if self._heartbeat and self._ts_activity - self._ts_heartbeat > Heartbeat.OFFLINE_TIMEOUT:
            _logger.info("%r: Much more recent activity than the last heartbeat, we've gone zombie", self)
            self._heartbeat = None

        online = (ts - max(self._ts_heartbeat, self._ts_activity)) <= Heartbeat.OFFLINE_TIMEOUT
        port_introspection_valid = (ts - self._ts_port_list) <= PortList.MAX_PUBLICATION_PERIOD * 2

        return NodeState(
            online=online,
            heartbeat=self._heartbeat,
            info=self._info,
            ports=self._ports if port_introspection_valid else None,
        )

    def __repr__(self) -> str:
        return str(pyuavcan.util.repr_attributes(self, node_id=self._node_id))


def expand_subjects(m: uavcan.node.port.SubjectIDList_0_1) -> AbstractSet[int]:
    if m.sparse_list is not None:
        return frozenset(int(x.value) for x in m.sparse_list)
    if m.mask:
        return expand_mask(m.mask)
    if m.total:
        return _COMPLETE_SUBJECT_SET
    assert False


def expand_mask(mask: NDArray[np.bool_]) -> AbstractSet[int]:
    return frozenset(x for x in range(len(mask)) if mask[x])


N_NODES = 65535  # The theoretical limit for all kinds of transports.
N_SUBJECTS = MessageDataSpecifier.SUBJECT_ID_MASK + 1
N_SERVICES = ServiceDataSpecifier.SERVICE_ID_MASK + 1

_COMPLETE_SUBJECT_SET = frozenset(range(N_SUBJECTS))
"""Made static for performance reasons."""

_logger = yakut.get_logger(__name__)



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import sys
import time
import enum
import itertools
import dataclasses
from concurrent.futures import Future, ThreadPoolExecutor, wait
from typing import Dict, List, Optional, Callable, Any, Sequence
from pathlib import Path
import logging
from ._child import Child
from ._schema import Composition, load_composition, load_ast, SchemaError
from ._schema import Statement, ShellStatement, CompositionStatement, JoinStatement


FlagDelegate = Callable[[], bool]


class ErrorCode(enum.IntEnum):
    """
    POSIX systems can safely use exit codes in [0, 125]. We use the upper range for own errors.
    https://unix.stackexchange.com/questions/418784/what-is-the-min-and-max-values-of-exit-codes-in-linux
    """

    SCHEMA_ERROR = 125
    FILE_ERROR = 124


@dataclasses.dataclass(frozen=True)
class Context:
    lookup_paths: Sequence[Path]
    poll_interval: float = 0.05


def locate(ctx: Context, file: str) -> Optional[Path]:
    p = Path(file)
    if p.is_absolute():
        if p.exists():
            return p
    else:
        for p in ctx.lookup_paths:
            p = (p / file).resolve()
            if p.exists():
                return p
    return None


def exec_file(
    ctx: Context, file: str, inout_env: Dict[str, bytes], *, gate: FlagDelegate, stack: Optional[Stack] = None
) -> int:
    """
    This function never raises exceptions in response to invalid syntax or a programmable error.
    Instead, it uses exit codes to report failures, to unify behavior with invoked processes.
    An exception may only indicate a bug in the implementation or an internal contract violation.

    The env var dict is used both as the input and the output.
    The provided values are inherited by the executed composition.
    Afterwards, they are updated with the variables defined by the composition, which take precedence
    over the supplied variables.
    """
    stack = stack or Stack()
    stack.log_debug(f"Locating file: {file!r} in:", *map(str, ctx.lookup_paths))
    pth = locate(ctx, file)
    if not pth:
        stack.log_warning(f"Cannot locate file {file!r} in:", *map(str, ctx.lookup_paths))
        return int(ErrorCode.FILE_ERROR)

    stack.log_debug(f"Executing file {file!r} found at: {pth}")
    try:
        source_text = pth.read_text()
    except Exception as ex:  # pylint: disable=broad-except
        stack.log_warning(f"Cannot read file {pth}: {ex}")
        return int(ErrorCode.FILE_ERROR)

    try:
        comp = load_composition(load_ast(source_text), inout_env.copy())
    except SchemaError as ex:
        stack.log_warning(f"Cannot load file {pth}: {ex}")
        return int(ErrorCode.SCHEMA_ERROR)

    # Export the variables to the caller. Vars from the composition override the supplied vars.
    inout_env.update(comp.env)

    stack = stack.push(repr(file))
    stack.log_debug(f"Loaded composition:", str(comp))
    return exec_composition(ctx, comp, gate=gate, stack=stack)


def exec_composition(ctx: Context, comp: Composition, *, gate: FlagDelegate, stack: Stack) -> int:
    env = comp.env.copy()
    for ca in comp.ext:  # The "env" is updated in-place.
        res = exec_file(ctx, ca.file, env, gate=gate, stack=stack.push("external"))
        if res != 0:
            return res

    def scr(node: str, scr: Sequence[Statement], inner_gate: FlagDelegate) -> int:
        inner_stack = stack.push(node)
        started_at = time.monotonic()
        res = exec_script(ctx, scr, env.copy(), kill_timeout=comp.kill_timeout, gate=inner_gate, stack=inner_stack)
        elapsed = time.monotonic() - started_at
        inner_stack.log_debug(f"Script exit status {res} in {elapsed:.1f} sec")
        return res

    res = scr("?", comp.predicate, gate)
    if res != 0:
        return 0
    res = scr("$", comp.main, gate)  # The return code of a composition is that of the first failed process.
    res_fin = scr(".", comp.fin, lambda: True)
    return res if res != 0 else res_fin


def exec_script(
    ctx: Context,
    scr: Sequence[Statement],
    env: Dict[str, bytes],
    *,
    kill_timeout: float,
    gate: FlagDelegate,
    stack: Stack,
) -> int:
    """
    :return: Exit code of the first statement to fail. Zero if all have succeeded.
    """
    if not scr:
        return 0  # We have successfully done nothing. Hard to fail that.

    first_failure_code: Optional[int] = None

    def inner_gate() -> bool:
        return (first_failure_code is None) and gate()

    def accept_result(result: int) -> None:
        nonlocal first_failure_code
        assert isinstance(result, int)
        if result != 0 and first_failure_code is None:
            first_failure_code = result  # Script ALWAYS returns the code of the FIRST FAILED statement.

    def launch_shell(inner_stack: Stack, cmd: str) -> Future[None]:
        return executor.submit(
            lambda: accept_result(
                exec_shell(ctx, cmd, env.copy(), kill_timeout=kill_timeout, gate=inner_gate, stack=inner_stack)
            )
        )

    def launch_composition(inner_stack: Stack, comp: Composition) -> Future[None]:
        return executor.submit(lambda: accept_result(exec_composition(ctx, comp, gate=inner_gate, stack=inner_stack)))

    executor = ThreadPoolExecutor(max_workers=len(scr))
    pending: List[Future[None]] = []
    try:
        for index, stmt in enumerate(scr):
            stmt_stack = stack.push(index)
            if not inner_gate():
                break
            if isinstance(stmt, ShellStatement):
                pending.append(launch_shell(stmt_stack, stmt.cmd))
            elif isinstance(stmt, CompositionStatement):
                pending.append(launch_composition(stmt_stack, stmt.comp))
            elif isinstance(stmt, JoinStatement):
                num_pending = sum(1 for x in pending if not x.done())
                stmt_stack.log_debug(f"Waiting for {num_pending} pending statements to join")
                if pending:
                    wait(pending)
            else:
                assert False

        # Wait for all statements to complete and then aggregate the results.
        done, not_done = wait(pending)
        assert not not_done
        _ = list(x.result() for x in done)  # Collect results explicitly to propagate exceptions.
        if first_failure_code is not None:
            assert first_failure_code != 0
            return first_failure_code
        return 0
    except Exception:
        first_failure_code = 1
        raise


def exec_shell(
    ctx: Context, cmd: str, env: Dict[str, bytes], *, kill_timeout: float, gate: FlagDelegate, stack: Stack
) -> int:
    started_at = time.monotonic()
    ch = Child(cmd, env, stdout=sys.stdout.buffer, stderr=sys.stderr.buffer)
    prefix = f"PID={ch.pid:08d} "
    try:
        longest_env = max(map(len, env.keys())) if env else 0
        stack.log_info(
            *itertools.chain(
                (f"{prefix}EXECUTING WITH ENVIRONMENT VARIABLES:",),
                ((k.ljust(longest_env) + " = " + repr(v.decode("raw_unicode_escape"))) for k, v in env.items())
                if env
                else ["<no variables>"],
                cmd.splitlines(),
            ),
        )
        ret: Optional[int] = None
        while gate() and ret is None:
            ret = ch.poll(ctx.poll_interval)
        if ret is None:
            stack.log_warning(f"{prefix}Stopping (was started {time.monotonic() - started_at:.1f} sec ago)")
            ch.stop(kill_timeout * 0.5, kill_timeout)
        while ret is None:
            ret = ch.poll(ctx.poll_interval)

        elapsed = time.monotonic() - started_at
        stack.log_info(f"{prefix}Exit status {ret} in {elapsed:.1f} sec")
        return ret
    finally:
        ch.kill()


class Stack:
    def __init__(self, path: Optional[List[str]] = None, logger: Optional[logging.Logger] = None) -> None:
        from . import __name__ as nm

        self._path = path or []
        self._logger = logger or logging.getLogger(nm)

    def push(self, node: Any) -> Stack:
        if isinstance(node, Path):
            node = repr(str(node))
        else:
            node = str(node)
        assert isinstance(node, str)
        return Stack(self._path + [node], self._logger)

    def log(self, level: int, *lines: str) -> None:
        if self._logger.isEnabledFor(level):
            self._logger.log(level, "Call stack: %s\n%s", self, "\n".join(lines))

    def log_debug(self, *lines: str) -> None:
        return self.log(logging.DEBUG, *lines)

    def log_info(self, *lines: str) -> None:
        return self.log(logging.INFO, *lines)

    def log_warning(self, *lines: str) -> None:
        return self.log(logging.WARNING, *lines)

    def __str__(self) -> str:
        return " ".join(self._path)



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import dataclasses
from typing import Dict, Sequence, Any, List
from yakut.yaml import Loader
from ._env import encode, flatten_registers, NAME_SEP, EnvironmentVariableError


NOT_ENV = "="
"""
Equals sign is the only character that cannot occur in an environment variable name in most OS.
"""


class SchemaError(ValueError):
    pass


@dataclasses.dataclass(frozen=True)
class Composition:
    env: Dict[str, bytes]
    ext: Sequence[External]

    predicate: Sequence[Statement]
    main: Sequence[Statement]
    fin: Sequence[Statement]

    @property
    def kill_timeout(self) -> float:
        try:
            # This is very tentative and is not yet specified. May be changed.
            return float(self.env.get("(kill_timeout)"))  # type: ignore
        except (ValueError, TypeError):
            return 20.0


@dataclasses.dataclass(frozen=True)
class Statement:
    pass


@dataclasses.dataclass(frozen=True)
class ShellStatement(Statement):
    cmd: str


@dataclasses.dataclass(frozen=True)
class CompositionStatement(Statement):
    comp: Composition


@dataclasses.dataclass(frozen=True)
class JoinStatement(Statement):
    pass


@dataclasses.dataclass(frozen=True)
class External:
    file: str


def load_ast(text: str) -> Any:
    try:
        return Loader().load(text)
    except Exception as ex:
        raise SchemaError(f"Syntax error: {ex}") from ex


def load_composition(ast: Any, env: Dict[str, bytes]) -> Composition:
    """
    Environment inheritance order (last entry takes precedence):

    - Parent process environment (i.e., the environment the orchestrator is invoked from).
    - Outer composition environment (e.g., root members of the orchestration file).
    - Local environment variables.
    """
    if not isinstance(ast, dict):
        raise SchemaError(f"The composition shall be a dict, not {type(ast).__name__}")
    ast = ast.copy()  # Prevent mutation of the outer object.
    env = env.copy()  # Prevent mutation of the outer object.
    try:
        for name, value in flatten_registers(
            {k: v for k, v in ast.items() if isinstance(k, str) and NOT_ENV not in k}
        ).items():
            if NAME_SEP in name:  # UAVCAN register.
                name = name.upper().replace(NAME_SEP, "_" * 2)
            if value is not None:
                env[name] = encode(value)
            else:
                env.pop(name, None)  # None is used to unset variables.
    except EnvironmentVariableError as ex:
        raise SchemaError(f"Environment variable error: {ex}") from EnvironmentVariableError

    out = Composition(
        env=env.copy(),
        ext=load_external(ast.pop("external=", [])),
        predicate=load_script(ast.pop("?=", []), env.copy()),
        main=load_script(ast.pop("$=", []), env.copy()),
        fin=load_script(ast.pop(".=", []), env.copy()),
    )
    unattended = [k for k in ast if NOT_ENV in k]
    if unattended:
        raise SchemaError(f"Unknown directives: {unattended}")
    return out


def load_script(ast: Any, env: Dict[str, bytes]) -> Sequence[Statement]:
    if isinstance(ast, list):
        return [load_statement(x, env) for x in ast]
    return [load_statement(ast, env)]


def load_statement(ast: Any, env: Dict[str, bytes]) -> Statement:
    if isinstance(ast, str):
        return ShellStatement(ast)
    if isinstance(ast, dict):
        return CompositionStatement(load_composition(ast, env))
    if ast is None:
        return JoinStatement()
    raise SchemaError("Statement shall be either: string (command to run), dict (nested schema), null (join)")


def load_external(ast: Any) -> List[External]:
    def item(inner: Any) -> External:
        if isinstance(inner, str):
            return External(inner)
        raise SchemaError(f"Call arguments shall be strings, not {type(ast).__name__}")

    if isinstance(ast, list):
        return [item(x) for x in ast]
    return [item(ast)]



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import math
from typing import Tuple, List, Sequence, Callable, Any, Dict
import dataclasses
import logging
import textwrap
import click
import pyuavcan
import yakut
from yakut.helpers import EnumParam
from yakut.yaml import EvaluableLoader
from yakut.util import construct_port_id_and_type
from ._executor import Executor, Publication

_MIN_SEND_TIMEOUT = 0.1
"""
With a slow garbage-collected language like Python, having a smaller timeout does not make practical sense.
This may be made configurable later.
"""


@dataclasses.dataclass(frozen=True)
class ExpressionContextModule:
    """
    Defines a Python module that is imported and available for use within YAML-embedded user expressions.
    """

    module_name: str
    docs_uri: str
    alias: str = ""
    wildcard: bool = False

    def __str__(self) -> str:
        if self.wildcard:
            body = f"from {self.module_name} import *"
        elif self.alias:
            body = f"import {self.module_name} as {self.alias}"
        else:
            body = f"import {self.module_name}"
        return body.ljust(25) + f"# {self.docs_uri}"

    @staticmethod
    def load(what: Sequence[ExpressionContextModule]) -> Dict[str, Any]:
        import inspect

        out: Dict[str, Any] = {}
        for ecm in what:
            _logger.debug("Importing: %r", ecm)
            mod = __import__(ecm.module_name)
            out[ecm.module_name] = mod
            if ecm.alias:
                out[ecm.alias] = mod
            if ecm.wildcard:
                out.update({name: member for name, member in inspect.getmembers(mod) if not name.startswith("_")})

        _logger.debug("Expression context contains %d items (listed on the next line):\n%s", len(out), list(out))
        return out


_EXPRESSION_CONTEXT_MODULES = [
    ExpressionContextModule("random", "https://docs.python.org/library/random.html", wildcard=True),
    ExpressionContextModule("time", "https://docs.python.org/library/time.html", wildcard=True),
    ExpressionContextModule("math", "https://docs.python.org/library/math.html", wildcard=True),
    ExpressionContextModule("os", "https://docs.python.org/library/os.html"),
    ExpressionContextModule("pyuavcan", "https://pyuavcan.readthedocs.io"),
    ExpressionContextModule("numpy", "https://numpy.org", alias="np"),
    ExpressionContextModule("scipy.constants", "https://docs.scipy.org/doc/scipy/reference/constants.html"),
    ExpressionContextModule("scipy.interpolate", "https://docs.scipy.org/doc/scipy/reference/interpolate.html"),
    ExpressionContextModule("scipy.linalg", "https://docs.scipy.org/doc/scipy/reference/linalg.html"),
    ExpressionContextModule("scipy.spatial", "https://docs.scipy.org/doc/scipy/reference/spatial.html"),
]

_EXAMPLES = """
Example: publish constant messages (no embedded expressions, just regular YAML):

\b
    yakut pub uavcan.diagnostic.Record.1.1 '{text: "Hello world!", severity: {value: 4}}' -N3 -T0.1 -P hi
    yakut pub 33:uavcan/si/unit/angle/Scalar_1_0 'radian: 2.31' uavcan.diagnostic.Record.1.1 'text: "2.31 rad"'

Example: publish sinewave with frequency 1 Hz, amplitude 10 meters:

\b
    yakut pub -T 0.01 1234:uavcan.si.unit.length.Scalar.1.0 '{meter: !$ "sin(t * pi * 2) * 10"}'

Example: as above, but control the frequency of the sinewave and its amplitude using sliders 10 and 11
of the first connected controller (use `yakut joystick` to find connected controllers and their axis mappings):

\b
    yakut pub -T 0.01 1234:uavcan.si.unit.length.Scalar.1.0 '{meter: !$ "sin(t * pi * 2 * A(1,10)) * 10 * A(1,11)"}'

Example: publish 3D angular velocity setpoint, thrust setpoint, and the arming switch state;
use positional initialization instead of YAML dicts:

\b
    yakut pub -T 0.1 \\
        5:uavcan.si.unit.angular_velocity.Vector3.1.0 '!$ "[A(1,0)*10, A(1,1)*10, (A(1,2)-A(1,5))*5]"' \\
        6:uavcan.si.unit.power.Scalar.1.0 '!$ A(2,10)*1e3' \\
        7:uavcan.primitive.scalar.Bit.1.0 '!$ T(1,5)'

Example: simulate timestamped measurement of voltage affected by white noise with standard deviation 0.25 V:

\b
    yakut pub -T 0.1 6:uavcan.si.sample.voltage.Scalar.1.0 \\
        '{timestamp: !$ time()*1e6, volt: !$ "A(2,10)*100+normalvariate(0,0.25)"}'
""".strip()

_HELP = f"""
Publish messages on the specified subjects.
Unless the local transport is configured in anonymous node,
the local node will also publish on standard subjects like Heartbeat and provide some standard RPC-services
like GetInfo.

The command accepts a list of space-separated pairs like:

\b
    [SUBJECT_ID:]TYPE_NAME.MAJOR.MINOR  YAML_FIELDS

The first element is a name like `uavcan.node.Heartbeat.1.0` prepended by the subject-ID.
The subject-ID may be omitted if a fixed one is defined for the data type.

The second element specifies the values of the message fields in YAML format (or JSON, which is a subset of YAML).
Missing fields will be left at their default values.
For more info about the format see DSDL API docs at https://pyuavcan.readthedocs.io.

The number of such pairs can be arbitrary; all defined messages will be published synchronously.
If no such pairs are specified, only the heartbeat will be published, unless the local node is anonymous.

The YAML document may embed arbitrary Python expressions that are re-evaluated immediately before publication.
Such expressions are annotated with the YAML tag `!$` (exclamation followed by dollar).
The result of such expression is substituted into the original YAML structure;
as such, the result can be of arbitrary type as long as the final YAML structure can be applied to the specified
DSDL instance.

The YAML-embedded expressions have access to the following variables (type is specified after the colon):

{Executor.SYM_INDEX}: int --- index of the current publication cycle, zero initially.

{Executor.SYM_TIME}: float --- time elapsed since first message (t=n*period).

{Executor.SYM_DTYPE}: Type[pyuavcan.dsdl.CompositeType] --- message class.

{Executor.SYM_CTRL_AXIS}: (controller,axis:int)->float ---
read the normalized value of `axis` from `controller` (e.g., joystick or MIDI fader).
To see the list of available controllers and determine their channel mapping, refer to `yakut joystick --help`.

{Executor.SYM_CTRL_BUTTON}: (controller,button:int)->bool ---
read the state of `button` from `controller` (true while held down).

{Executor.SYM_CTRL_TOGGLE}: (controller,toggle:int)->bool ---
read the state of `toggle` from `controller`.

The following Python modules are imported and usable within embedded expressions;
refer to their API docs for usage info:

\b
{textwrap.indent(chr(0x0A).join(map(str, _EXPRESSION_CONTEXT_MODULES)), " " * 4)}

{_EXAMPLES}
"""


def _validate_message_spec(
    ctx: click.Context,
    param: click.Parameter,
    value: Tuple[str, ...],
) -> List[Tuple[str, str]]:
    if len(value) % 2 != 0:
        raise click.BadParameter(
            f"Message specifier shall have an even number of paired arguments (found {len(value)} arguments)",
            ctx=ctx,
            param=param,
        )
    return [(s, f) for s, f in (value[i : i + 2] for i in range(0, len(value), 2))]  # pylint: disable=R1721


@yakut.subcommand(help=_HELP)
@click.argument(
    "message",
    type=str,
    callback=_validate_message_spec,
    metavar="SUBJECT FIELDS [SUBJECT FIELDS]...",
    nargs=-1,
)
@click.option(
    "--period",
    "-T",
    type=float,
    default=1.0,
    show_default=True,
    metavar="SECONDS",
    help=f"""
Message publication period.
All messages are published synchronously, so the period setting applies to all specified subjects.
The send timeout equals the period as long as it is not less than {_MIN_SEND_TIMEOUT} seconds.
""",
)
@click.option(
    "--count",
    "-N",
    type=int,
    default=2 ** 64 - 1,
    metavar="CARDINAL",
    help=f"Number of publication cycles before exiting normally. Unlimited by default.",
)
@click.option(
    "--priority",
    "-P",
    default=pyuavcan.presentation.DEFAULT_PRIORITY,
    type=EnumParam(pyuavcan.transport.Priority),
    help=f"Priority of published message transfers. [default: {pyuavcan.presentation.DEFAULT_PRIORITY.name}]",
)
@yakut.pass_purser
@yakut.asynchronous
async def publish(
    purser: yakut.Purser,
    message: Sequence[Tuple[str, str]],
    period: float,
    count: int,
    priority: pyuavcan.transport.Priority,
) -> None:
    _logger.debug("period=%s, count=%s, priority=%s, message=%s", period, count, priority, message)
    assert all((isinstance(a, str) and isinstance(b, str)) for a, b in message)
    assert isinstance(period, float) and isinstance(count, int) and isinstance(priority, pyuavcan.transport.Priority)
    if period < 1e-9 or not math.isfinite(period):
        raise click.BadParameter("Period shall be a positive real number of seconds")
    if count <= 0:
        _logger.info("Nothing to do because count=%s", count)
        return

    send_timeout = max(_MIN_SEND_TIMEOUT, period)
    loader = EvaluableLoader(ExpressionContextModule.load(_EXPRESSION_CONTEXT_MODULES))

    def make_publication_factory(
        subject_spec: str, field_spec: str
    ) -> Callable[[pyuavcan.presentation.Presentation], Publication]:
        subject_id, dtype = construct_port_id_and_type(subject_spec)
        # Catch errors as early as possible.
        if issubclass(dtype, pyuavcan.dsdl.ServiceObject):
            raise click.BadParameter(f"Subject spec {subject_spec!r} refers to a service type")
        # noinspection PyTypeChecker
        if subject_id is None and pyuavcan.dsdl.get_fixed_port_id(dtype) is None:
            raise click.UsageError(
                f"Subject-ID is not provided and {pyuavcan.dsdl.get_model(dtype)} does not have a fixed one"
            )
        try:
            evaluator = loader.load_unevaluated(field_spec)
        except ValueError as ex:
            raise click.BadParameter(f"Invalid field spec {field_spec!r}: {ex}") from None
        _logger.debug("Publication spec appears valid: %r", subject_spec)
        return lambda presentation: Publication(
            subject_id=subject_id,
            dtype=dtype,
            evaluator=evaluator,
            presentation=presentation,
            priority=priority,
            send_timeout=send_timeout,
        )

    # This is to perform as much processing as possible before constructing the node.
    # Catching errors early allows us to avoid disturbing the network and peripherals unnecessarily.
    publication_factories = [make_publication_factory(*m) for m in message]

    node = purser.get_node("publish", allow_anonymous=True)
    executor = Executor(
        node=node,
        loader=loader,
        publications=(f(node.presentation) for f in publication_factories),
    )
    try:  # Even if the publication set is empty, we still have to publish the heartbeat.
        _logger.info(
            "Publishing %d subjects with period %.3fs, send timeout %.3fs, count %d, priority %s",
            len(publication_factories),
            period,
            send_timeout,
            count,
            priority.name,
        )
        await executor.run(count=count, period=period)
    finally:
        executor.close()
        if _logger.isEnabledFor(logging.INFO):
            _logger.info("%s", node.presentation.transport.sample_statistics())
            for s in node.presentation.transport.output_sessions:
                ds = s.specifier.data_specifier
                if isinstance(ds, pyuavcan.transport.MessageDataSpecifier):
                    _logger.info("Subject %d: %s", ds.subject_id, s.sample_statistics())


_logger = yakut.get_logger(__name__)



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import asyncio
from typing import TYPE_CHECKING, Callable, Any, Type
import threading
import pyuavcan
from pyuavcan.transport import Timestamp, AlienTransfer
import yakut

if TYPE_CHECKING:
    import pyuavcan.application  # pylint: disable=ungrouped-imports


class Iface:
    def __init__(self, node: pyuavcan.application.Node) -> None:
        self._loop = asyncio.get_event_loop()
        self._node = node
        self._clients: dict[
            tuple[Type[pyuavcan.dsdl.CompositeObject], int],
            pyuavcan.presentation.Client[pyuavcan.dsdl.CompositeObject],
        ] = {}
        self._subscriptions: list[Type[pyuavcan.dsdl.FixedPortCompositeObject]] = []
        self._lock = threading.RLock()
        self._trace_handlers: list[Callable[[Timestamp, AlienTransfer], None]] = []
        self._transport_error_handlers: list[Callable[[pyuavcan.transport.ErrorTrace], None]] = []
        self._tracer = node.presentation.transport.make_tracer()

        _logger.info("Starting packet capture on %r", self._node)
        node.presentation.transport.begin_capture(self._process_capture)

    def add_trace_handler(self, cb: Callable[[Timestamp, AlienTransfer], None]) -> None:
        self._trace_handlers.append(cb)

    def add_transport_error_handler(self, cb: Callable[[pyuavcan.transport.ErrorTrace], None]) -> None:
        self._transport_error_handlers.append(cb)

    def add_standard_subscription(self, dtype: Type[pyuavcan.dsdl.FixedPortCompositeObject]) -> None:
        """
        It is necessary to create subscriptions even if we're not going to use them directly to ensure that
        the network is informed about our interest in this data.
        For UAVCAN/CAN and UAVCAN/serial this may not matter but for UAVCAN/UDP this is important
        because we need to let the underlying layers publish the relevant IGMP states.
        """
        if dtype not in self._subscriptions:  # This is just to avoid excessive subscriptions.
            _logger.info("Subscribing to the fixed subject-ID of %r", dtype.__name__)
            self._node.make_subscriber(dtype).receive_in_background(self._dummy_subscription_handler)
            self._subscriptions.append(dtype)

    def try_request(
        self,
        dtype: Type[pyuavcan.dsdl.FixedPortServiceObject],
        server_node_id: int,
        request: pyuavcan.dsdl.CompositeObject,
    ) -> None:
        """
        The expectation is that the response will be read via the packet capture interface.
        If the local node is anonymous, nothing would happen.
        The request is always executed at the lowest priority level.
        """
        if self._node.id is None or self._node.id == server_node_id:
            return
        _logger.info("Requesting fixed service-ID of %r at %r", dtype.__name__, server_node_id)
        try:
            client = self._clients[dtype, server_node_id]
        except LookupError:
            self._clients[dtype, server_node_id] = self._node.make_client(dtype, server_node_id)
            client = self._clients[dtype, server_node_id]
            client.priority = pyuavcan.transport.Priority.OPTIONAL
            client.response_timeout = 5.0

        async def run() -> None:
            _ = await client.call(request)

        self._loop.create_task(run())

    def _process_capture(self, cap: pyuavcan.transport.Capture) -> None:
        # Locking is super critical! Captures may run in separate threads.
        with self._lock:
            trace = self._tracer.update(cap)
            if isinstance(trace, pyuavcan.transport.TransferTrace):
                self._loop.call_soon_threadsafe(
                    pyuavcan.util.broadcast(self._trace_handlers), trace.timestamp, trace.transfer
                )
            elif isinstance(trace, pyuavcan.transport.ErrorTrace):
                self._loop.call_soon_threadsafe(pyuavcan.util.broadcast(self._transport_error_handlers), trace)
            else:
                pass

    @staticmethod
    async def _dummy_subscription_handler(*_: Any) -> None:
        pass


_logger = yakut.get_logger(__name__)



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

# pylint: disable=too-many-locals

from __future__ import annotations
import sys
import functools
from typing import TYPE_CHECKING, Optional, Callable, Any, AbstractSet, TypeVar
from collections import defaultdict
import numpy as np
import pyuavcan
import yakut
from ._model import N_NODES, N_SUBJECTS, N_SERVICES, NodeState
from ._ui import Style, Color, Canvas, TableRenderer

if TYPE_CHECKING:
    from numpy.typing import NDArray
    from scipy.sparse import spmatrix
    import uavcan.node


S_DEFAULT = Style(fg=Color.WHITE, salience=1)
S_FAILURE = Style(fg=Color.WHITE, bg=Color.RED, salience=2)
S_MUTED = Style(salience=-1)
S_ADVISORY = Style(fg=Color.MAGENTA, salience=1)
S_CAUTION = Style(fg=Color.YELLOW, salience=1)
S_WARNING = Style(fg=Color.RED, salience=1)
S_NOTICE = Style(fg=Color.CYAN, salience=1)

S_NICE = Style(fg=Color.GREEN, salience=1)
S_POOR = Style(fg=Color.YELLOW, salience=1)


class View:
    _CONNECTIVITY_MATRIX_CELL_WIDTH = 5

    def __init__(self) -> None:
        self._fragments: list[str] = []

        self._node_table_renderer = TableRenderer(map(len, View._NODE_TABLE_HEADER), separate_columns=True)
        self._connectivity_matrix_renderer = TableRenderer(
            (View._CONNECTIVITY_MATRIX_CELL_WIDTH for _ in range(N_NODES + 1)),
            separate_columns=False,
        )

        legend_canvas = Canvas()
        row = 0
        col = legend_canvas.put(row, 0, "APPLICATION LAYER CONNECTIVITY MATRIX [t/s=transfer/second] Colors: ")
        col = legend_canvas.put(
            row, col, "pub/cln", style=get_matrix_cell_style(tx=True, rx=False, recently_active=False)
        )
        col = legend_canvas.put(row, col, "│")
        col = legend_canvas.put(
            row, col, "sub/srv", style=get_matrix_cell_style(tx=False, rx=True, recently_active=False)
        )
        col = legend_canvas.put(row, col, "│")
        col = legend_canvas.put(
            row,
            col,
            "(pub+sub)/(cln+srv)",
            style=get_matrix_cell_style(tx=True, rx=True, recently_active=False),
        )
        col = legend_canvas.put(row, col, "│")
        col = legend_canvas.put(
            row, col, "activity", style=get_matrix_cell_style(tx=False, rx=False, recently_active=True)
        )
        col = legend_canvas.put(row, col, "│uavcan.node.port.List is ")
        col = legend_canvas.put(row, col, "published", style=S_NICE)
        col = legend_canvas.put(row, col, "/")
        col = legend_canvas.put(row, col, "not", style=S_POOR)
        legend_canvas.put(row, col, "│")

        self._connectivity_matrix_legend = legend_canvas.flip_buffer()

    def flip_buffer(self) -> str:
        out = "\n".join(self._fragments)
        self._fragments = []
        return out

    def render(
        self,
        states: dict[Optional[int], NodeState],
        xfer_deltas: spmatrix,
        xfer_rates: spmatrix,
        byte_rates: spmatrix,
        total_transport_errors: int,
        fir_window_duration: float,
    ) -> None:
        self._fragments.append(self._render_node_table(states))

        self._fragments.append(self._connectivity_matrix_legend)
        self._fragments.append(self._render_connectivity_matrix(states, xfer_deltas, xfer_rates, byte_rates))

        annotation_canvas = Canvas()
        col = annotation_canvas.put(0, 0, "Total transport layer errors:")
        col = annotation_canvas.put(
            0,
            col,
            f"{total_transport_errors:9d}",
            style=S_POOR if total_transport_errors > 0 else S_NICE,
        )
        col += 9
        col = annotation_canvas.put(0, col, f"Values averaged over {fir_window_duration:.1f} sec")
        _ = col
        self._fragments.append(annotation_canvas.flip_buffer())

    # noinspection SpellCheckingInspection
    _NODE_TABLE_HEADER = [
        "NodID",
        "Mode",
        "Health",
        "VSSC",
        "Uptime".ljust(14),
        "VProtcl",
        "VHardwr",
        "VSoftware(major.minor.vcs.crc)".ljust(41),
        "Unique-ID".ljust(32),
        "Name",
    ]

    def _render_node_table(self, states: dict[Optional[int], NodeState]) -> str:
        for idx, s in enumerate(View._NODE_TABLE_HEADER):
            self._node_table_renderer[0, idx] = s

        for idx, (node_id, ss) in enumerate(states.items()):
            col = 0

            def put(data: Any, style: Optional[Style]) -> None:
                nonlocal col
                self._node_table_renderer[idx + 1, col] = data, (  # pylint: disable=cell-var-from-loop
                    style or S_DEFAULT
                )
                col += 1

            if node_id is not None:
                if not ss.online:
                    put(node_id, S_MUTED)
                else:
                    put(node_id, None)
            else:
                put("anon", None if ss.online else S_MUTED)

            if ss.heartbeat:
                if node_id is None and ss.online:
                    sty_override: Optional[Style] = S_FAILURE  # Anonymous nodes shall not publish heartbeat.
                elif not ss.online:
                    sty_override = S_MUTED
                else:
                    sty_override = None
                txt, sty = render_mode(ss.heartbeat.mode)
                put(txt, sty_override or sty)
                txt, sty = render_health(ss.heartbeat.health)
                put(txt, sty_override or sty)
                put(ss.heartbeat.vendor_specific_status_code, sty_override)
            else:
                put("?", S_MUTED)
                put("?", S_MUTED)
                put("?", S_MUTED)

            if ss.online:
                if ss.heartbeat is None:
                    if node_id is not None:
                        put("zombie", S_FAILURE)
                    else:
                        put("online", None)
                else:
                    put(render_uptime(ss.heartbeat.uptime), None)
            else:
                put("offline", S_MUTED)

            if ss.info:
                sty = None if ss.online and ss.heartbeat else S_MUTED
                put(
                    render_version(ss.info.protocol_version),
                    sty if ss.info.protocol_version.major == pyuavcan.UAVCAN_SPECIFICATION_VERSION[0] else S_FAILURE,
                )
                put(render_version(ss.info.hardware_version), sty)
                put(
                    render_full_software_version(
                        ss.info.software_version,
                        ss.info.software_vcs_revision_id,
                        int(ss.info.software_image_crc[0]) if len(ss.info.software_image_crc) > 0 else None,
                    ),
                    sty,
                )
                put(ss.info.unique_id.tobytes().hex(), sty)
                # Best effort to display bad names
                put("".join(ss.info.name.tobytes().decode(errors="ignore").split()), sty)
            else:
                for _ in range(5):
                    put("?", S_MUTED)

        return self._node_table_renderer.flip_buffer()

    # noinspection SpellCheckingInspection
    def _render_connectivity_matrix(
        self,
        states: dict[Optional[int], NodeState],
        xfer_delta: spmatrix,
        xfer_rates: spmatrix,
        byte_rates: spmatrix,
    ) -> str:
        tbl = self._connectivity_matrix_renderer
        online_states: dict[Optional[int], NodeState] = {k: v for k, v in states.items() if v.online}

        # This part took some time to get right to avoid accidental dense matrix operations, which are super slow.
        xfer_rates_by_ds = xfer_rates.sum(axis=0)
        assert xfer_rates_by_ds.size == N_SUBJECTS + N_SERVICES * 2
        xfer_delta_by_ds = xfer_delta.sum(axis=0)
        byte_rates_by_ds = byte_rates.sum(axis=0)

        # Consider a port existing if either holds:
        #   - there have been recent transfers, even if the source nodes have gone offline
        #   - if the port was recently reported via uavcan.node.port.List, even if the node is currently offline
        all_subjects: set[int] = set()
        all_services: set[int] = set()
        for y in xfer_rates_by_ds.nonzero()[1]:
            y = int(y)
            if y < N_SUBJECTS:
                all_subjects.add(y)
            else:
                all_services.add((y - N_SUBJECTS) % N_SERVICES)
        for node_id, state in states.items():
            if state.ports is not None:
                all_subjects |= state.ports.pub
                # Subjects that are only subscribed to by supersubscribers are only shown if there are other nodes
                # utilizing these.
                if len(state.ports.sub) < N_SUBJECTS:
                    all_subjects |= state.ports.sub
                all_services |= state.ports.cln
                all_services |= state.ports.srv

        # HEADER ROWS AND COLUMNS
        num_nodes = len(online_states)
        num_subjects = len(all_subjects)
        num_services = len(all_services)
        row_subject = 0
        row_service = row_subject + num_subjects + 3
        row_total = row_service + num_services + 3

        tbl[row_subject, 0] = "MESSG", S_MUTED
        tbl[row_service, 0] = "RQ+RS", S_MUTED
        tbl[row_total, 0] = "TOTAL", S_MUTED

        for row in (row_subject + num_subjects + 1, row_service + num_services + 1):
            tbl[row + 0, num_nodes + 3] = "↖ t/s", S_MUTED
            tbl[row + 1, num_nodes + 3] = "", S_MUTED

        for row in (row_subject, row_service, row_total):  # Row of node-IDs and per-port totals.
            for ii, (node_id, state) in enumerate(online_states.items()):
                sty = S_POOR if state.ports is None else S_NICE
                if node_id is not None:
                    tbl[row, ii + 1] = node_id, sty
                else:
                    tbl[row, ii + 1] = " anon", sty
            tbl[row, num_nodes + 1] = " ∑t/s"
            tbl[row, num_nodes + 2] = " ∑B/s"

        for row in (row_subject + num_subjects, row_service + num_services, row_total):  # Per-node totals.
            tbl[row + 1, 0] = "∑t/s"
            tbl[row + 2, 0] = "∑B/s"

        for ii, sid in enumerate(sorted(all_subjects)):  # Subject-ID and Service-ID.
            for col in (0, num_nodes + 3):
                tbl[row_subject + ii + 1, col] = sid, S_DEFAULT
        for ii, sid in enumerate(sorted(all_services)):
            for col in (0, num_nodes + 3):
                tbl[row_service + ii + 1, col] = sid, S_DEFAULT

        # CONTENTS
        View._render_subject_matrix_contents(
            lambda row, col, data, style: tbl.set_cell(row + row_subject + 1, col + 1, data, style=style),
            states=online_states,
            subjects=all_subjects,
            xfer_delta=xfer_delta,
            xfer_rates=xfer_rates,
            byte_rates=byte_rates,
            xfer_delta_by_port=xfer_delta_by_ds,
            xfer_rates_by_port=xfer_rates_by_ds,
            byte_rates_by_port=byte_rates_by_ds,
        )

        def slice_req_rsp(m: _T) -> tuple[_T, _T]:
            a = N_SUBJECTS + N_SERVICES * 0
            b = N_SUBJECTS + N_SERVICES * 1
            c = N_SUBJECTS + N_SERVICES * 2
            return (m[:, a:b], m[:, b:c])  # type: ignore

        View._render_service_matrix_contents(
            lambda row, col, data, style: tbl.set_cell(row + row_service + 1, col + 1, data, style=style),
            states=online_states,
            services=all_services,
            xfer_delta=slice_req_rsp(xfer_delta),
            xfer_rates=slice_req_rsp(xfer_rates),
            byte_rates=slice_req_rsp(byte_rates),
            xfer_delta_by_port=slice_req_rsp(xfer_delta_by_ds),
            xfer_rates_by_port=slice_req_rsp(xfer_rates_by_ds),
            byte_rates_by_port=slice_req_rsp(byte_rates_by_ds),
        )

        # TOTAL DATA RATE
        xfer_delta_by_node = xfer_delta.sum(axis=1)
        xfer_rates_by_node = xfer_rates.sum(axis=1)
        byte_rates_by_node = byte_rates.sum(axis=1)
        for ii, node_id in enumerate(online_states):
            x = node_id if node_id is not None else N_NODES
            sty = get_matrix_cell_style(None, None, int(xfer_delta_by_node[x]) > 0)
            tbl[row_total + 1, ii + 1] = render_xfer_rate(float(xfer_rates_by_node[x])), sty
            tbl[row_total + 2, ii + 1] = render_xfer_rate(float(byte_rates_by_node[x])), sty

        # Sum the DS-wise vectors because they are usually faster due to being smaller.
        sty = get_matrix_cell_style(None, None, int(xfer_delta_by_ds.sum()) > 0)
        tbl[row_total + 1, num_nodes + 1] = render_xfer_rate(float(xfer_rates_by_ds.sum())), sty
        tbl[row_total + 2, num_nodes + 2] = render_byte_rate(float(byte_rates_by_ds.sum())), sty

        tbl[row_total + 1, num_nodes + 3] = ""
        tbl[row_total + 2, num_nodes + 3] = ""

        return tbl.flip_buffer()

    @staticmethod
    def _render_subject_matrix_contents(
        put: Callable[[int, int, Any, Style], None],
        states: dict[Optional[int], NodeState],
        subjects: AbstractSet[int],
        xfer_delta: spmatrix,
        xfer_rates: spmatrix,
        byte_rates: spmatrix,
        xfer_delta_by_port: NDArray[np.int_],
        xfer_rates_by_port: NDArray[np.float_],
        byte_rates_by_port: NDArray[np.float_],
    ) -> None:
        recent_by_node: dict[Optional[int], bool] = defaultdict(bool)
        xfer_rate_by_node: dict[Optional[int], float] = defaultdict(float)
        byte_rate_by_node: dict[Optional[int], float] = defaultdict(float)
        for row, subject_id in enumerate(sorted(subjects)):
            for col, (node_id, state) in enumerate(states.items()):
                x = node_id if node_id is not None else N_NODES
                recent = int(xfer_delta[x, subject_id]) > 0
                rate = float(xfer_rates[x, subject_id])
                if state.ports is not None:
                    pub = subject_id in state.ports.pub
                    sub = subject_id in state.ports.sub
                    sty = get_matrix_cell_style(pub, sub, recent)
                    text = render_xfer_rate(rate) if pub or (rate > EPSILON) else ""
                else:
                    sty = get_matrix_cell_style(None, None, recent)
                    text = render_xfer_rate(rate) if rate > EPSILON else ""
                put(row, col, text, sty)

                recent_by_node[node_id] |= recent
                xfer_rate_by_node[node_id] += rate
                byte_rate_by_node[node_id] += byte_rates[x, subject_id]

            recent = xfer_delta_by_port[0, subject_id] > 0
            sty = get_matrix_cell_style(None, None, recent)
            put(row, len(states) + 0, render_xfer_rate(xfer_rates_by_port[0, subject_id]), sty)
            put(row, len(states) + 1, render_byte_rate(byte_rates_by_port[0, subject_id]), sty)

        row = len(subjects)
        for col, node_id in enumerate(states):
            sty = get_matrix_cell_style(None, None, recent_by_node[node_id])
            put(row + 0, col, render_xfer_rate(xfer_rate_by_node[node_id]), sty)
            put(row + 1, col, render_byte_rate(byte_rate_by_node[node_id]), sty)

        sty = get_matrix_cell_style(None, None, sum(recent_by_node.values()) > 0)
        put(row + 0, len(states) + 0, render_xfer_rate(sum(xfer_rate_by_node.values())), sty)
        put(row + 1, len(states) + 1, render_byte_rate(sum(byte_rate_by_node.values())), sty)

    @staticmethod
    def _render_service_matrix_contents(
        put: Callable[[int, int, Any, Style], None],
        states: dict[Optional[int], NodeState],
        services: AbstractSet[int],
        xfer_delta: tuple[spmatrix, spmatrix],
        xfer_rates: tuple[spmatrix, spmatrix],
        byte_rates: tuple[spmatrix, spmatrix],
        xfer_delta_by_port: tuple[NDArray[np.int_], NDArray[np.int_]],
        xfer_rates_by_port: tuple[NDArray[np.float_], NDArray[np.float_]],
        byte_rates_by_port: tuple[NDArray[np.float_], NDArray[np.float_]],
    ) -> None:
        # We used to display two rows per service: separate request and response. It is very informative but a bit
        # expensive in terms of the screen space, which is very limited when large networks are involved.
        # So while the data provided to this method is sufficient to build a super-detailed representation,
        # currently we collapse it into one service per row such that request and response states are joined together.
        # We may change it later shall the need arise.
        xfer_delta_uni: spmatrix = sum(xfer_delta)
        byte_rates_uni: spmatrix = sum(byte_rates)
        xfer_delta_by_port_uni: NDArray[np.int_] = sum(xfer_delta_by_port)  # type: ignore
        xfer_rates_by_port_uni: NDArray[np.float_] = sum(xfer_rates_by_port)  # type: ignore
        byte_rates_by_port_uni: NDArray[np.float_] = sum(byte_rates_by_port)  # type: ignore

        recent_by_node: dict[Optional[int], bool] = defaultdict(bool)
        xfer_rate_by_node: dict[Optional[int], float] = defaultdict(float)
        byte_rate_by_node: dict[Optional[int], float] = defaultdict(float)

        for row, service_id in enumerate(sorted(services)):
            for col, (node_id, state) in enumerate(states.items()):
                x = node_id if node_id is not None else N_NODES
                recent = int(xfer_delta_uni[x, service_id]) > 0
                rate_req = float(xfer_rates[0][x, service_id])
                rate_rsp = float(xfer_rates[1][x, service_id])
                rate_total = rate_req + rate_rsp
                if state.ports is not None:
                    cln = service_id in state.ports.cln
                    srv = service_id in state.ports.srv
                    sty = get_matrix_cell_style(tx=cln, rx=srv, recently_active=recent)
                    text = render_xfer_rate(rate_total) if cln or srv or (rate_total > EPSILON) else ""
                else:
                    sty = get_matrix_cell_style(None, None, recent)
                    text = render_xfer_rate(rate_total) if rate_total > EPSILON else ""

                put(row, col, text, sty)

                recent_by_node[node_id] |= recent
                xfer_rate_by_node[node_id] += rate_total
                byte_rate_by_node[node_id] += byte_rates_uni[x, service_id]

            recent = int(xfer_delta_by_port_uni[0, service_id]) > 0
            sty = get_matrix_cell_style(None, None, recent)
            put(row, len(states) + 0, render_xfer_rate(xfer_rates_by_port_uni[0, service_id]), sty)
            put(row, len(states) + 1, render_byte_rate(byte_rates_by_port_uni[0, service_id]), sty)

        total_recent = False
        total_xfer_rate = 0.0
        total_byte_rate = 0.0

        row = len(services)
        for col, node_id in enumerate(states):
            recent = recent_by_node[node_id] > 0
            xfer_rate = xfer_rate_by_node[node_id]
            byte_rate = byte_rate_by_node[node_id]

            total_recent = total_recent or recent
            total_xfer_rate += xfer_rate
            total_byte_rate += byte_rate

            sty = get_matrix_cell_style(None, None, recent)
            put(row + 0, col, render_xfer_rate(xfer_rate), sty)
            put(row + 1, col, render_byte_rate(byte_rate), sty)

        sty = get_matrix_cell_style(None, None, total_recent)
        put(row + 0, len(states) + 0, render_xfer_rate(total_xfer_rate), sty)
        put(row + 1, len(states) + 1, render_byte_rate(total_byte_rate), sty)


@functools.lru_cache(None)
def get_matrix_cell_style(tx: Optional[bool], rx: Optional[bool], recently_active: bool) -> Style:
    salience = 1 if recently_active else -1
    fg = Color.RED if recently_active else Color.WHITE
    if tx and rx:
        return Style(fg=fg, bg=Color.CYAN, salience=salience)
    if tx:
        return Style(fg=fg, bg=Color.BLUE, salience=salience)
    if rx:
        return Style(fg=fg, bg=Color.GREEN, salience=salience)
    return Style(fg=fg, salience=salience)


# noinspection SpellCheckingInspection
def render_mode(val: uavcan.node.Mode_1_0) -> tuple[str, Optional[Style]]:
    if val.value == val.OPERATIONAL:
        return "oper", None
    if val.value == val.INITIALIZATION:
        return "init", S_NOTICE
    if val.value == val.MAINTENANCE:
        return "mntn", S_ADVISORY
    if val.value == val.SOFTWARE_UPDATE:
        return "swup", S_CAUTION
    return str(val.value), S_FAILURE  # pragma: no cover


# noinspection SpellCheckingInspection
def render_health(val: uavcan.node.Health_1_0) -> tuple[str, Optional[Style]]:
    if val.value == val.NOMINAL:
        return "nomina", None
    if val.value == val.ADVISORY:
        return "adviso", S_ADVISORY
    if val.value == val.CAUTION:
        return "cautio", S_CAUTION
    if val.value == val.WARNING:
        return "warnin", S_WARNING
    return str(val.value), S_FAILURE  # pragma: no cover


def render_uptime(val: int) -> str:
    return f"{val // (3600 * 24):5d}d{(val // 3600) % 24:02d}:{(val // 60) % 60:02d}:{val % 60:02d}"


def render_version(val: uavcan.node.Version_1_0) -> str:
    return "% 3d.%-3d" % (val.major, val.minor)  # pylint: disable=consider-using-f-string


def render_full_software_version(version: uavcan.node.Version_1_0, vcs_revision_id: int, crc: Optional[int]) -> str:
    out = f"{version.major:3d}.{version.minor}"
    if vcs_revision_id != 0 or crc is not None:
        out += f".{vcs_revision_id:016x}"
    if crc is not None:
        out += f".{crc:016x}"
    return out.ljust(41)


def render_xfer_rate(x: float) -> str:
    x = max(x, 0.0)  # The value may be slightly negative due to accumulated floating point error
    if x < 1e3:
        return f"{x:4.0f} "
    if x < 1e6:
        return f"{x / 1e3:4.0f}k"
    return f"{x / 1e6:4.0f}M"


def render_byte_rate(x: float) -> str:
    x = max(x, 0.0)  # The value may be slightly negative due to accumulated floating point error
    if x < 1024:
        return f"{x:4.0f} "
    if x < 1024 * 1024:
        return f"{x / 1024:4.0f}K"
    return f"{x / (1024 * 1024):4.0f}M"


EPSILON = sys.float_info.epsilon

_T = TypeVar("_T")

_logger = yakut.get_logger(__name__)



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from typing import Any
import decimal
import ruamel.yaml
import ruamel.yaml.constructor


class Loader:
    """
    YAML parsing facade.
    """

    def __init__(self) -> None:
        self._impl = ruamel.yaml.YAML()

    def load(self, text: str) -> Any:
        return self._impl.load(text)


def _unittest_yaml() -> None:
    import pytest
    from ._dumper import Dumper

    ref = Dumper(explicit_start=True).dumps(
        {
            "abc": decimal.Decimal("-inf"),
            "def": [decimal.Decimal("nan"), {"qaz": decimal.Decimal("789")}],
        }
    )
    assert Loader().load(ref) == {
        "abc": -float("inf"),
        "def": [pytest.approx(float("nan"), nan_ok=True), {"qaz": pytest.approx(789)}],
    }



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import abc
import sys
from typing import Iterable, Tuple, Callable
import dataclasses
import pyuavcan.util
import yakut

if sys.version_info >= (3, 9):
    from collections.abc import Mapping
else:  # pragma: no cover
    from typing import Mapping  # pylint: disable=ungrouped-imports


class ControllerError(RuntimeError):
    """
    Base class for controller-related errors.
    """


class ControllerNotFoundError(ControllerError):
    """
    Controller could not be opened or sampled because it is no longer connected or never was.
    """


class Controller(abc.ABC):
    """
    Models a human interface controller like a joystick or a MIDI controller.
    The access is normally non-exclusive, meaning that multiple processes can access the same device concurrently,
    unless this is not supported by the hardware or the platform.
    """

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """
        Human-readable name of this controller. Examples:

        - ``FAD.9:FAD.9 MIDI 1 24:0``
        - ``Xbox 360 Controller``

        This property does not raise exceptions.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def sample(self) -> Sample:
        """
        Sample the current state of the controls and return it.

        :raises: :class:`ControllerError`
        """
        raise NotImplementedError

    @abc.abstractmethod
    def set_update_hook(self, hook: Callable[[], None]) -> None:
        """
        The update hook is invoked shortly after the controller state is updated, possibly from a different thread.
        If a hook is already installed, the behavior is undefined.
        Invoking :meth:`sample` from the hook may or may not be supported.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        """
        Dispose the instance and the underlying resources. This method is idempotent.
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.name!r})"

    @staticmethod
    def list_controllers() -> Iterable[Tuple[str, Callable[[], Controller]]]:
        """
        List devices that are usable with this specific class.
        Normally you should use the freestanding function :func:`list_controllers`.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class Sample:
    """
    An atomic sample of the current controller's state.
    """

    axis: Mapping[int, float]
    """
    Analogue axes whose values are normalized into [0, +1] for unipolar channels and [-1, +1] for bipolar channels.
    """

    button: Mapping[int, bool]
    """
    Each push button channel is True while the button is held down by the user, False otherwise.
    """

    toggle: Mapping[int, bool]
    """
    Toggle switch channels alternate their state between True and False at each user interaction.
    """


def list_controllers() -> Iterable[Tuple[str, Callable[[], Controller]]]:
    """
    Use this function for listing available devices and constructing new instances of :class:`Controller`.
    The :class:`null.NullController` is always available and listed first (i.e., at index 0, as the name suggests).

    :return: Iterable of tuples of (unique name, factory), one per available device.
    """
    from .null import NullController

    def handle_import_error(module_name: str, culprit: Exception) -> None:
        _logger.warning(
            "Could not import controller module %r; controllers of this kind may not be usable. Error: %s",
            module_name,
            str(culprit) or repr(culprit),
        )

    pyuavcan.util.import_submodules(sys.modules[__name__], handle_import_error)
    base = Controller
    # Order controller kinds by class name, but ensure that NullController always comes first.
    for ty in sorted(
        pyuavcan.util.iter_descendants(base),
        key=lambda x: (x is not NullController, x.__name__),
    ):
        try:
            options = list(ty.list_controllers())
        except Exception as ex:  # pylint: disable=broad-except
            _logger.warning("Could not list controllers of kind %r: %s", ty, ex)
            options = []
        for name, factory in options:
            _logger.debug("Detected controller from %s: %r", ty.__name__, name)
            yield name, factory


_logger = yakut.get_logger(__name__)



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import os
import sys
import time
import signal
from subprocess import Popen, DEVNULL
from typing import Dict, Optional, Callable, List, Tuple, BinaryIO, Any
import yakut

# Disable unused ignore warning for this file only because there appears to be no other way to make MyPy
# accept this file both on Windows and GNU/Linux.
# mypy: warn_unused_ignores=False

_logger = yakut.get_logger(__name__)


if sys.platform.startswith("win"):  # pragma: no cover
    SIGNAL_INTERRUPT = signal.CTRL_BREAK_EVENT
    SIGNAL_TERMINATE = signal.SIGTERM
    SIGNAL_KILL = signal.SIGTERM
else:
    SIGNAL_INTERRUPT = signal.SIGINT
    SIGNAL_TERMINATE = signal.SIGTERM
    SIGNAL_KILL = signal.SIGKILL


class Child:
    """
    Starts a child shell process and provides convenient non-blocking controls:

    - :meth:`poll` for querying the status.
    - :meth:`stop` for stopping with automatic escalation of the signal strength.
    - The constructor accepts file objects where the child process output is sent to.

    When running on Windows, it may be desirable to use PowerShell as the system shell.
    For that, export the COMSPEC environment variable.
    """

    def __init__(self, cmd: str, env: Dict[str, bytes], *, stdout: BinaryIO, stderr: BinaryIO) -> None:
        """
        :param cmd: Shell command to execute. Execution starts immediately.
        :param env: Additional environment variables.
        :param stdout: Stdout from the child process is redirected there.
        :param stderr: Ditto, but for stderr.
        """
        self._return: Optional[int] = None
        self._signaling_schedule: List[Tuple[float, Callable[[], None]]] = []
        e: Any
        if os.supports_bytes_environ:
            e = os.environb.copy()  # type: ignore
            e.update({k.encode(): v for k, v in env.items()})
        else:  # pragma: no cover
            e = os.environ.copy()
            e.update({k: v.decode() for k, v in env.items()})
        self._proc = Popen(  # pylint: disable=consider-using-with
            cmd,
            env=e,
            shell=True,
            stdout=stdout,
            stderr=stderr,
            stdin=DEVNULL,
            bufsize=1,
        )

    @property
    def pid(self) -> int:
        """
        The process-ID of the child. This value retains validity even after the child is terminated.
        """
        return self._proc.pid

    def poll(self, timeout: float) -> Optional[int]:
        """
        :param timeout: Block for this many seconds, at most.
        :return: None if still running, exit code if finished (idempotent).
        """
        if self._return is None:
            if self._signaling_schedule:
                deadline, handler = self._signaling_schedule[0]
                if time.monotonic() >= deadline:
                    self._signaling_schedule.pop(0)
                    handler()

            ret = self._proc.poll()
            if ret is None:
                time.sleep(timeout)
                ret = self._proc.poll()
            if ret is not None:
                self._return = ret

        return self._return

    def stop(self, escalate_after: float, give_up_after: float) -> None:
        """
        Send a SIGINT/CTRL_BREAK_EVENT to the process and schedule to check if it's dead later.

        :param escalate_after: If the process is still alive this many seconds after the initial termination signal,
            send a SIGTERM.

        :param give_up_after: Ditto, but instead of SIGTERM send SIGKILL (on Windows reuse SIGTERM instead)
            and disown the child immediately without waiting around. This is logged as error.
        """
        if self._return is not None or self._proc.poll() is not None:
            return
        give_up_after = max(give_up_after, escalate_after)
        _logger.debug(
            "%s: Stopping using signal %r. Escalation timeout: %.1f, give-up timeout: %.1f",
            self,
            SIGNAL_INTERRUPT,
            escalate_after,
            give_up_after,
        )
        signal_tree(self.pid, SIGNAL_INTERRUPT)

        def terminate() -> None:
            _logger.warning("%s: The child is still alive. Escalating to %r", self, SIGNAL_TERMINATE)
            signal_tree(self.pid, SIGNAL_TERMINATE)

        def kill() -> None:
            _logger.error(
                "%s: The child is still alive. Escalating to %r and detaching. No further attempts will be made!",
                self,
                SIGNAL_KILL,
            )
            self.kill()

        now = time.monotonic()
        self._signaling_schedule = [
            (now + escalate_after, terminate),
            (now + give_up_after, kill),
        ]

    def kill(self) -> None:
        """
        This is intended for abnormal termination of the owner of this instance.
        Simply kills the child with all its children and ceases all related activities.
        """
        if self._return is None:
            self._return = -SIGNAL_KILL
        signal_tree(self.pid, SIGNAL_KILL)

    def __str__(self) -> str:
        return f"Child {self.pid:08d}"


def signal_tree(pid: int, sig: int) -> None:
    """
    Send the signal to the specified process and all its children. The parent is signaled first.
    The list of children is collected once before the parent is signaled.

    This is needed for properly terminating process hierarchies, particularly those launched by shells,
    because merely killing the parent does not necessarily terminate its children.
    If there is no such process, does nothing.
    """
    import psutil  # type: ignore

    try:
        parent = psutil.Process(pid)
    except psutil.NoSuchProcess:
        _logger.debug("No process with PID=%r", pid)
    else:
        children = parent.children()
        _logger.debug("Sending signal %r to process %r with direct children %r", sig, pid, [x.pid for x in children])
        parent.send_signal(sig)
        for c in children:
            signal_tree(c.pid, sig)


def _unittest_child(caplog: object) -> None:
    import pytest
    import logging
    from pathlib import Path

    assert isinstance(caplog, pytest.LogCaptureFixture)

    io_out = Path("stdout")
    io_err = Path("stderr")

    if not sys.platform.startswith("win"):
        py = (
            "import time, signal as s; "
            + "s.signal(s.SIGINT, lambda *_: None); "
            + "s.signal(s.SIGTERM, lambda *_: None); "
            + "time.sleep(10)"
        )

        with caplog.at_level(logging.CRITICAL):
            c = Child(f'python -c "{py}"', {}, stdout=io_out.open("wb"), stderr=io_err.open("wb"))
            assert c.poll(0.1) is None
            c.stop(1.0, 2.0)
            assert c.poll(1.0) is None
            for _ in range(50):
                res = c.poll(0.1)
            assert res is not None
            assert res < 0  # Killed
            c.stop(1.0, 2.0)  # No effect because already dead.
        assert not io_out.read_text()
        assert not io_err.read_text()

        c = Child(f"sleep 10", {}, stdout=io_out.open("wb"), stderr=io_err.open("wb"))
        assert c.poll(0.1) is None
        c.kill()
        res = c.poll(1.0)
        assert res is not None
        assert res < 0  # Killed
        assert not io_out.read_text()
        assert not io_err.read_text()

        c = Child(
            """python -c "import sys; print('ABC', file=sys.stderr); print('DEF'); print('GHI', end='')" """,
            {},
            stdout=io_out.open("wb"),
            stderr=io_err.open("wb"),
        )
        assert 0 == c.poll(1.0)
        time.sleep(2.0)
        assert io_out.read_text().splitlines() == ["DEF", "GHI"]
        assert io_err.read_text().splitlines() == ["ABC"]

    else:  # pragma: no cover
        c = Child(f"echo Hello", {}, stdout=io_out.open("wb"), stderr=io_err.open("wb"))
        assert 0 == c.poll(5.0)
        c.stop(2.0, 4.0)
        c.kill()
        assert io_out.read_text().splitlines() == ["Hello"]
        assert not io_err.read_text()

    io_out.unlink(missing_ok=True)
    io_err.unlink(missing_ok=True)
